import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist

import os
import math
import numpy as np
from transformers import AutoModel
import scipy.sparse as sp
import pandas as pd
from tqdm import tqdm

from utils.nns_utils import ExactSearch
from utils.dl_utils import create_tf_pooler, ToD, BatchIterator, DistBatchIterator, apply_and_accumulate, GatherLayer, RandContext
from utils.topk_utils import TopK


class BaseNet(nn.Module):
    def __init__(self):
        super().__init__()

    def ToD(self, batch):
        return ToD(batch, self.get_device())

    def get_device(self):
        if hasattr(self, 'device'):
            return self.device
        return list(self.parameters())[0].device

    def get_embs(self, data_source, bsz=256, accelerator=None, encode_func=None):
        self.eval()
        if isinstance(data_source, torch.utils.data.Dataset): 
            data_source = BatchIterator(data_source, bsz) if accelerator is None else DistBatchIterator(data_source, bsz)

        encode_func = self.encode if encode_func is None else encode_func 
        
        out = apply_and_accumulate(
            data_source, 
            lambda b: {'embs': encode_func(self.ToD(b))},
            accelerator,
            display_name='Embedding'
            )
        return out['embs'] if 'embs' in out else None

    def _predict_batch(self, b, K):
        b = ToD(b, self.get_device())
        out = self(b)
        if isinstance(out, torch.Tensor): # BxL shaped out
            top_vals, top_inds = torch.topk(out, K)
        elif isinstance(out, tuple) and len(out) == 2: # (logits, indices) shaped out 
            top_vals, temp_inds = torch.topk(out[0], K)
            top_inds = torch.gather(out[1], 1, temp_inds)
        return {'top_vals': top_vals, 'top_inds': top_inds}

    def predict(self, data_source, K=100, bsz=256, accelerator=None):
        self.eval()
        if isinstance(data_source, torch.utils.data.Dataset):
            data_source = BatchIterator(data_source, bsz)

        out = apply_and_accumulate(
            data_source, 
            self._predict_batch,
            accelerator,
            display_name='Predicting',
            **{'K': K}
            )

        if accelerator is None or accelerator.is_main_process:
            labels = data_source.dataset.labels
            indptr = np.arange(0, labels.shape[0]*K+1, K)
            score_mat = sp.csr_matrix((out['top_vals'].ravel(), out['top_inds'].ravel(), indptr), labels.shape)
            # remove padding if any
            if any(score_mat.indices == labels.shape[1]):
                score_mat.data[score_mat.indices == labels.shape[1]] = 0
                score_mat.eliminate_zeros()
            return score_mat

    def update_non_parameters(self, *args, **kwargs):
        pass

    def save(self, path):
        torch.save(self.state_dict(), path)

    def load(self, path):
        loaded_state = torch.load(path, map_location='cpu')
        return self.load_state_dict(loaded_state, strict=False)

class TFEncoder(BaseNet):
    def __init__(self, args):
        super().__init__()
        tf_args = {'add_pooling_layer': False} if any(args.tf.startswith(name) for name in ['bert-base', 'roberta-base', 'sentence', 'BAAI']) else {}
        self.tf = AutoModel.from_pretrained(args.tf, **tf_args) if args.tf else None
        self.tf_pooler, self.tf_dims = create_tf_pooler(args.tf_pooler)
        self.tf_dims = self.tf.config.hidden_size
        self.bottleneck = nn.Linear(self.tf_dims, args.bottleneck_dim) if args.bottleneck_dim else None
        self.embs_dim = args.embs_dim = args.bottleneck_dim if args.bottleneck_dim else self.tf_dims
        self.dropout = nn.Dropout(args.dropout)
        self.norm_embs = args.norm_embs
        self.amp_encode = args.amp_encode

    def encode(self, b, encoder=None, pooler=None):
        with torch.cuda.amp.autocast(self.amp_encode):
            embs = b['xfts']
            if encoder is None:
                encoder = self.tf
            if pooler is None:
                pooler = self.tf_pooler
            embs = pooler(encoder(**embs, output_hidden_states=True), embs)
            if self.bottleneck is not None:
                embs = self.bottleneck(embs)

            embs = self.dropout(embs)
            if self.norm_embs:
                embs = F.normalize(embs, dim=-1)
            return embs.float()

class DistBaseNetwork(TFEncoder):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args)
        self.accelerator = accelerator
        if accelerator is not None:
            self.rank = accelerator.state.process_index
            self.world_size = accelerator.state.num_processes
        else:
            self.rank = 0
            self.world_size = 1
        self.numy = args.numy
        self.numy_after_pad = int(self.world_size * np.ceil(self.numy / self.world_size))
        self.local_y_inds = torch.arange(self.rank*self.numy_after_pad//self.world_size, min((self.rank+1)*self.numy_after_pad//self.world_size, self.numy))
        self.tau = args.tau
        print(f"Rank {self.rank} of {self.world_size} initialized")

    def encode(self, b, encoder=None, pooler=None):
        return super().encode(b, encoder, pooler).contiguous()

    def compute_loss(self, sim, targets, loss_fn):
        @torch.no_grad()
        def filter_non_local_yinds(rows, cols, vals=None):
            local_y_mask = (cols >= self.local_y_inds[0]) & (cols < self.local_y_inds[-1])
            if vals is None: return rows[local_y_mask], cols[local_y_mask] - self.local_y_inds[0]
            else: return rows[local_y_mask], cols[local_y_mask] - self.local_y_inds[0], vals[local_y_mask]

        with self.accelerator.no_sync(self):
            return loss_fn(sim, targets, filter_non_local_yinds)

    def ignore_labels(self, sim, ignore_labels):
        ignore_labels = np.intersect1d(self.local_y_inds, ignore_labels)
        if len(ignore_labels) > 0:
            ignore_labels -= self.local_y_inds[0].numpy()
            sim[:,ignore_labels] = -np.inf
        return sim


class DistClassifierNetwork(DistBaseNetwork):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        self.local_y_inds = self.local_y_inds.to(f'cuda:{self.rank}')
        self.local_w = nn.Linear(self.embs_dim, self.numy_after_pad//self.world_size)

    def forward_head(self, xembs):
        return self.local_w(xembs)
    
    def forward_backward(self, b, loss_fn, scaler=None, ignore_labels=None):
        # forward
        with torch.cuda.amp.autocast(enabled=scaler is not None):
            local_xembs = self.encode(b)
            all_xembs = GatherLayer.apply(local_xembs).view(-1, *local_xembs.shape[1:])

        with torch.no_grad():
            max_pad_len = self.accelerator.gather([torch.tensor(b['y']['inds'].shape[1], device=self.get_device())])[0].max().item()
            b['y']['inds'] = F.pad(b['y']['inds'], (0, max_pad_len - b['y']['inds'].shape[1]), value=self.numy)
            b['y']['vals'] = F.pad(b['y']['vals'], (0, max_pad_len - b['y']['vals'].shape[1]), value=0)
            target_inds, target_vals = self.accelerator.gather([b['y']['inds'], b['y']['vals']])

        # with torch.cuda.amp.autocast(enabled=scaler is not None):
        sim = self.forward_head(all_xembs)
        sim /= self.tau 
        if self.rank == self.world_size-1 and self.numy_after_pad > self.numy:
            sim = sim[:, :-(self.numy_after_pad-self.numy)]

        if ignore_labels is not None:
            sim = self.ignore_labels(sim, ignore_labels)
                
        all_targets = (target_inds, target_vals)
        loss = self.compute_loss(sim, all_targets, loss_fn)

        with self.accelerator.no_sync(self):
            loss.backward() if scaler is None else scaler.scale(loss).backward()
            for name, param in self.named_parameters():
                if 'local_' not in name:
                    dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
                    # param.grad.data /= self.world_size
        dist.all_reduce(loss, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
        return loss
    
    def _predict_batch(self, b, K):
        b = ToD(b, self.get_device())
        xembs = self.encode(b)
        xinds, xembs = self.accelerator.gather([b['ids'], xembs])

        sim = self.forward_head(xembs)
        if self.rank == self.world_size-1 and self.numy_after_pad > self.numy:
            sim = sim[:, :-(self.numy_after_pad-self.numy)]

        topk_sim = torch.topk(sim, min(K, sim.shape[1]), dim=1)
        topk_inds = self.local_y_inds[topk_sim.indices]
        topk_vals = topk_sim.values
        del sim, xembs
        topk_inds, topk_vals = self.accelerator.gather([topk_inds.unsqueeze(0), topk_vals.unsqueeze(0)])
        topk_inds = topk_inds.permute(1, 0, 2).reshape(topk_inds.shape[1], -1)
        topk_vals = topk_vals.permute(1, 0, 2).reshape(topk_vals.shape[1], -1)
        topk_vals, topk_temp_inds = topk_vals.topk(K, dim=1)
        topk_inds = topk_inds.gather(1, topk_temp_inds)
        return xinds, topk_inds, topk_vals
    
    @torch.no_grad()
    def predict(self, data_loader, K=100, bsz=256, **kwargs):
        self.eval()
        if self.rank == 0:
            all_topk_inds = torch.zeros((len(data_loader.dataset), K), dtype=torch.long)
            all_topk_vals = torch.zeros((len(data_loader.dataset), K), dtype=torch.float)

        for b in tqdm(data_loader, desc='Predicting', disable=self.rank!=0):
            xinds, topk_inds, topk_vals = self._predict_batch(b, K)
            if self.rank == 0:
                all_topk_inds[xinds] = topk_inds.detach().cpu()
                all_topk_vals[xinds] = topk_vals.detach().cpu()
        
        if self.rank == 0:
            labels = data_loader.dataset.labels
            indptr = np.arange(0, labels.shape[0]*K+1, K)
            score_mat = sp.csr_matrix((all_topk_vals.ravel(), all_topk_inds.ravel(), indptr), labels.shape)
            # remove padding if any
            if any(score_mat.indices == labels.shape[1]):
                score_mat.data[score_mat.indices == labels.shape[1]] = 0
                score_mat.eliminate_zeros()
            return score_mat
        
    def gather(self):
        if dist.is_initialized():
            all_w_weight = [torch.zeros_like(self.local_w.weight) for _ in range(dist.get_world_size())] if self.rank == 0 else None
            dist.gather(self.local_w.weight, all_w_weight, dst=0)
            all_w_bias = [torch.zeros_like(self.local_w.bias) for _ in range(dist.get_world_size())] if self.rank == 0 else None
            dist.gather(self.local_w.bias, all_w_bias, dst=0)
        else:
            all_w_weight = [self.local_w.weight]
            all_w_bias = [self.local_w.bias]
        if self.rank == 0:
            self.all_w_weight = torch.cat(all_w_weight, dim=0)
            self.all_w_bias = torch.cat(all_w_bias, dim=0)
        
    def save(self, path):
        state_dict = self.state_dict()
        state_dict = {n: p for n, p in state_dict.items() if 'local_w' not in n}
        state_dict['all_w.weight'] = self.all_w_weight
        state_dict['all_w.bias'] = self.all_w_bias
        torch.save(state_dict, path)

    def load(self, path):
        state_dict = torch.load(path, map_location=self.get_device())
        local_state_dict = {n: p for n, p in state_dict.items() if 'all_w' not in n}
        local_state_dict['local_w.weight'] = torch.zeros_like(self.local_w.weight)
        local_state_dict['local_w.bias'] = torch.zeros_like(self.local_w.bias)

        clf_chunk_size = self.numy_after_pad // self.world_size
        local_clf_inds = range(self.rank*clf_chunk_size, min((self.rank+1)*clf_chunk_size, self.numy))
        local_state_dict['local_w.weight'][:len(local_clf_inds)] = state_dict['all_w.weight'][local_clf_inds]
        local_state_dict['local_w.bias'][:len(local_clf_inds)] = state_dict['all_w.bias'][local_clf_inds]

        return self.load_state_dict(local_state_dict)


class DistLabelWiseAttentionNetwork(DistClassifierNetwork):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        self.local_attn = nn.Linear(self.embs_dim, self.numy_after_pad//self.world_size)

    def forward_head(self, xembs):
        attention = self.local_attn(xembs).transpose(1, 2) # (B, L, C) -> (B, numy, L)
        attention = F.softmax(attention, dim=-1)
        weighted_embs = torch.bmm(attention, xembs) # (B, numy, L) x (B, L, C) -> (B, numy, C)
        logits = (self.local_w.weight * weighted_embs).sum(dim=-1) + self.local_w.bias # (B, numy, C) -> (B, numy)
        return logits
    
    def gather(self):
        super().gather()
        if dist.is_initialized():
            all_attn_weight = [torch.zeros_like(self.local_attn.weight) for _ in range(dist.get_world_size())] if self.rank == 0 else None
            dist.gather(self.local_attn.weight, all_attn_weight, dst=0)
            all_attn_bias = [torch.zeros_like(self.local_attn.bias) for _ in range(dist.get_world_size())] if self.rank == 0 else None
            dist.gather(self.local_attn.bias, all_attn_bias, dst=0)
        else:
            all_attn_weight = [self.local_attn.weight]
            all_attn_bias = [self.local_attn.bias]
        if self.rank == 0:
            self.all_attn_weight = torch.cat(all_attn_weight, dim=0)
            self.all_attn_bias = torch.cat(all_attn_bias, dim=0)
        
    def save(self, path):
        state_dict = self.state_dict()
        state_dict = {n: p for n, p in state_dict.items() if 'local_' not in n}
        state_dict['all_w.weight'] = self.all_w_weight
        state_dict['all_w.bias'] = self.all_w_bias
        state_dict['all_attn.weight'] = self.all_attn_weight
        state_dict['all_attn.bias'] = self.all_attn_bias
        torch.save(state_dict, path)

    def load(self, path):
        state_dict = torch.load(path, map_location=self.get_device())
        local_state_dict = {n: p for n, p in state_dict.items() if 'all_' not in n}
        local_state_dict['local_w.weight'] = torch.zeros_like(self.local_w.weight)
        local_state_dict['local_w.bias'] = torch.zeros_like(self.local_w.bias)
        local_state_dict['local_attn.weight'] = torch.zeros_like(self.local_attn.weight)
        local_state_dict['local_attn.bias'] = torch.zeros_like(self.local_attn.bias)

        clf_chunk_size = self.numy_after_pad // self.world_size
        local_clf_inds = range(self.rank*clf_chunk_size, min((self.rank+1)*clf_chunk_size, self.numy))
        local_state_dict['local_w.weight'][:len(local_clf_inds)] = state_dict['all_w.weight'][local_clf_inds]
        local_state_dict['local_w.bias'][:len(local_clf_inds)] = state_dict['all_w.bias'][local_clf_inds]
        local_state_dict['local_attn.weight'][:len(local_clf_inds)] = state_dict['all_attn.weight'][local_clf_inds]
        local_state_dict['local_attn.bias'][:len(local_clf_inds)] = state_dict['all_attn.bias'][local_clf_inds]

        return self.load_state_dict(local_state_dict)

from collections import UserDict
class DistDualEncoderAll(DistBaseNetwork):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        self.gc_bsz = args.gc_bsz
        self.lbl_dataset = data_manager.lbl_dataset
        num_local_y = self.numy_after_pad//self.world_size
        if self.rank*num_local_y + self.local_y_inds.shape[0] > self.numy:
            self.local_y_inds = self.local_y_inds[:-(self.numy_after_pad - self.numy)]

    def get_input_tensors(self, model_input):
        """
        Recursively go through model input and grab all tensors, which are then used to record current device random
        states. This method will do its best to parse types of Tensor, tuple, list, dict and UserDict. Other types will
        be ignored unless self._get_input_tensors_strict is set to True, in which case an exception will be raised.
        :param model_input: input to model
        :return: all torch tensors in model_input
        """
        if isinstance(model_input, torch.Tensor):
            return [model_input]

        elif isinstance(model_input, (list, tuple)):
            return sum((self.get_input_tensors(x) for x in model_input), [])

        elif isinstance(model_input, (dict, UserDict)):
            return sum((self.get_input_tensors(x) for x in model_input.values()), [])

        elif self._get_input_tensors_strict:
            raise NotImplementedError(f'get_input_tensors not implemented for type {type(model_input)}')

        else:
            return []

    def encode(self, b, encoder=None, pooler=None):
        return super().encode(b, encoder=encoder, pooler=pooler).contiguous()
    
    def encode_x(self, b):
        return self.encode({'xfts': b['xfts']})

    def encode_y(self, b):
        return self.encode({'xfts': b['yfts'] if 'yfts' in b else b['xfts']})

    def local_encode_y(self, b):
        # return emb of all y and the random states
        rand_states = []
        inputs = []
        with torch.no_grad():
            concat_yembs = torch.zeros((self.local_y_inds.shape[0], self.embs_dim), device=self.get_device())
            for i in range(0, self.local_y_inds.shape[0], self.gc_bsz):
                input_range = range(i, min(i+self.gc_bsz, self.local_y_inds.shape[0]))
                yembs_input = self.ToD({'yfts': self.lbl_dataset.get_fts(self.local_y_inds[input_range])})
                inputs.append((yembs_input, input_range))
                rand_states.append(RandContext(*self.get_input_tensors(yembs_input)))
                concat_yembs[input_range] = self.encode_y(yembs_input)
        return inputs, concat_yembs, rand_states
    
    def forward_head(self, xembs, yembs, attn_mask):
        return xembs @ yembs.t()

    def backward_to_embs(self, loss, local_xembs, local_yembs, scaler=None):
        if scaler is not None:
            loss = scaler.scale(loss)
        local_xembs_grad, local_yembs_grad = torch.autograd.grad(loss, (local_xembs, local_yembs))
        return local_xembs_grad, local_yembs_grad
        
    def forward_backward(self, b, loss_fn, scaler=None, ignore_labels=None):
        # forward
        with torch.no_grad():
            max_pad_len = self.accelerator.gather([torch.tensor(b['y']['inds'].shape[1], device=self.get_device())])[0].max().item()
            b['y']['inds'] = F.pad(b['y']['inds'], (0, max_pad_len - b['y']['inds'].shape[1]), value=self.numy)
            b['y']['vals'] = F.pad(b['y']['vals'], (0, max_pad_len - b['y']['vals'].shape[1]), value=0)
            target_inds, target_vals = self.accelerator.gather([b['y']['inds'], b['y']['vals']])

            max_pad_len = self.accelerator.gather([torch.tensor(b['xfts']['attention_mask'].shape[1], device=self.get_device())])[0].max().item()
            b['xfts']['input_ids'] = F.pad(b['xfts']['input_ids'], (0, max_pad_len - b['xfts']['input_ids'].shape[1]), value=0)
            b['xfts']['attention_mask'] = F.pad(b['xfts']['attention_mask'], (0, max_pad_len - b['xfts']['attention_mask'].shape[1]), value=0)
            all_attn_mask = self.accelerator.gather(b['xfts']['attention_mask'])

        local_xembs = self.encode_x(b)
        all_xembs = GatherLayer.apply(local_xembs).view(-1, *local_xembs.shape[1:])
        local_yinputs, local_yembs, local_yembs_rand_states = self.local_encode_y(b)
        local_yembs.requires_grad = True


        sim = self.forward_head(all_xembs, local_yembs, all_attn_mask)
        sim /= self.tau
        if ignore_labels is not None:
            sim = self.ignore_labels(sim, ignore_labels)
        all_targets = (target_inds, target_vals)
        loss = self.compute_loss(sim, all_targets, loss_fn)

        # backward using gradient caching
        with self.accelerator.no_sync(self):
            local_xembs.grad, local_yembs.grad = self.backward_to_embs(loss, local_xembs, local_yembs, scaler)

            # do normal backward through xembs
            torch.autograd.backward(local_xembs, local_xembs.grad)
            
            # do backward through yembs using cached gradients
            for (yembs_input, yinds_range), rand_state in zip(local_yinputs, local_yembs_rand_states):
                with rand_state:
                    yembs = self.encode_y(yembs_input)
                yembs.grad = local_yembs.grad[yinds_range]
                torch.autograd.backward(yembs, yembs.grad)

            for param in self.parameters():
                dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
                param.grad.data /= self.world_size
        dist.all_reduce(loss, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
        return loss

    @torch.no_grad()
    def predict(self, data_loader, K = 100, bsz = 256, **kwargs):
        x_embs = self.get_embs(data_loader.dataset.x_dataset, bsz, encode_func=self.encode_x, accelerator=self.accelerator)
        y_embs = self.get_embs(data_loader.dataset.y_dataset, bsz, encode_func=self.encode_y, accelerator=self.accelerator)
        if self.accelerator and not self.accelerator.is_main_process:
            return None
        es = ExactSearch(y_embs, device=self.get_device(), K=K)
        score_mat = es.search(x_embs)
        return score_mat

class DistAsymmetricDualEncoder(DistDualEncoderAll):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        tf_args = {'add_pooling_layer': False} if any(args.tf.startswith(name) for name in ['bert-base', 'roberta-base', 'sentence', 'BAAI']) else {}
        self.label_tf = AutoModel.from_pretrained(args.tf, **tf_args)
    
    def encode_y(self, b):
        return self.encode({'xfts': b['yfts'] if 'yfts' in b else b['xfts']}, encoder=self.label_tf)


class DistAsymmetricDualEncoderAll(DistDualEncoderAll):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        self.yembs = np.load(args.lbl_embs)
        self._local_yembs = None
        self.out_proj = nn.Linear(self.yembs.shape[-1], self.embs_dim)
        self.ss_weight = args.ss_weight if args.self_supervised_type is not None else None
    
    @property
    def local_yembs(self):
        if self._local_yembs is None:
            self._local_yembs = torch.from_numpy(self.yembs[self.local_y_inds]).float().to(self.get_device())
        return self._local_yembs
    
    def forward_head(self, all_xembs, attn_mask):
        yembs = F.normalize(self.out_proj(self.dropout(self.local_yembs)), dim=-1)
        xembs = F.normalize(all_xembs, dim=-1)
        return xembs @ yembs.t()

    @torch.no_grad()
    def get_y_embs(self):
        self.eval()
        yembs = torch.from_numpy(self.yembs).float().to(self.get_device())
        yembs = F.normalize(self.out_proj(self.dropout(yembs)), dim=-1)
        return yembs.detach().cpu().numpy()
    
    def encode_ss_y(self, b):
        return self.encode({'xfts': b['ss_yfts']})

    def self_supervised_forward_backward(self, b, loss_fn, scaler=None):
        b['xfts'] = b['ss_yfts']
        b['y'] = b['ss_y']
        return self.forward_backward(b, loss_fn, scaler, ss_weight=self.ss_weight)

    def forward_backward(self, b, loss_fn, scaler=None, ss_weight=None, ignore_labels=None):
        with torch.no_grad():
            max_pad_len = self.accelerator.gather([torch.tensor(b['y']['inds'].shape[1], device=self.get_device())])[0].max().item()
            b['y']['inds'] = F.pad(b['y']['inds'], (0, max_pad_len - b['y']['inds'].shape[1]), value=self.numy)
            b['y']['vals'] = F.pad(b['y']['vals'], (0, max_pad_len - b['y']['vals'].shape[1]), value=0)
            target_inds, target_vals = self.accelerator.gather([b['y']['inds'], b['y']['vals']])

            max_pad_len = self.accelerator.gather([torch.tensor(b['xfts']['attention_mask'].shape[1], device=self.get_device())])[0].max().item()
            b['xfts']['input_ids'] = F.pad(b['xfts']['input_ids'], (0, max_pad_len - b['xfts']['input_ids'].shape[1]), value=0)
            b['xfts']['attention_mask'] = F.pad(b['xfts']['attention_mask'], (0, max_pad_len - b['xfts']['attention_mask'].shape[1]), value=0)
            all_attn_mask = self.accelerator.gather(b['xfts']['attention_mask'])

        # forward
        local_xembs = self.encode_x(b)
        all_xembs = GatherLayer.apply(local_xembs).view(-1, *local_xembs.shape[1:])
        sim = self.forward_head(all_xembs, attn_mask=all_attn_mask)
        sim /= self.tau
        if ignore_labels is not None:
            sim = self.ignore_labels(sim, ignore_labels)

        all_targets = (target_inds, target_vals)
        loss = self.compute_loss(sim, all_targets, loss_fn) * (1 if ss_weight is None else ss_weight)

        with self.accelerator.no_sync(self):
            loss.backward() if scaler is None else scaler.scale(loss).backward()
            for param in self.parameters():
                dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
                param.grad.data /= self.world_size
        dist.all_reduce(loss, op=dist.ReduceOp.SUM) if dist.is_initialized() else None
        return loss

    def _predict_batch(self, b, K, local_y_inds):
        b = ToD(b, self.get_device())
        xembs = self.encode(b)
        # max_pad_len = self.accelerator.gather([torch.tensor(b['xfts']['attention_mask'].shape[1], device=self.get_device())])[0].max().item()
        # b['xfts']['attention_mask'] = F.pad(b['xfts']['attention_mask'], (0, max_pad_len - b['xfts']['attention_mask'].shape[1]), value=0)
        xinds, xembs, attn_mask = self.accelerator.gather([b['index'], xembs, b['xfts']['attention_mask']])

        sim = self.forward_head(xembs, attn_mask)

        topk_sim = torch.topk(sim, min(K, sim.shape[1]), dim=1)
        topk_inds = local_y_inds[topk_sim.indices]
        topk_vals = topk_sim.values
        del sim, xembs
        topk_inds, topk_vals = self.accelerator.gather([topk_inds.unsqueeze(0), topk_vals.unsqueeze(0)])
        topk_inds = topk_inds.permute(1, 0, 2).reshape(topk_inds.shape[1], -1)
        topk_vals = topk_vals.permute(1, 0, 2).reshape(topk_vals.shape[1], -1)
        topk_vals, topk_temp_inds = topk_vals.topk(K, dim=1)
        topk_inds = topk_inds.gather(1, topk_temp_inds)
        return xinds, topk_inds, topk_vals
    
    @torch.no_grad()
    def predict(self, data_loader, K=100, bsz=256, **kwargs):
        self.eval()

        if self.rank == 0:
            all_topk_inds = torch.zeros((len(data_loader.dataset), K), dtype=torch.long)
            all_topk_vals = torch.zeros((len(data_loader.dataset), K), dtype=torch.float)

        local_y_inds = ToD(self.local_y_inds, self.get_device())
        for b in tqdm(data_loader, desc='Predicting', disable=self.rank!=0):
            b['xfts'] = data_loader.dataset.get_fts(b['index'], 'x')
            xinds, topk_inds, topk_vals = self._predict_batch(b, K, local_y_inds)
            if self.rank == 0:
                all_topk_inds[xinds] = topk_inds.detach().cpu()
                all_topk_vals[xinds] = topk_vals.detach().cpu()
        
        if self.rank == 0:
            labels = data_loader.dataset.labels
            indptr = np.arange(0, labels.shape[0]*K+1, K)
            score_mat = sp.csr_matrix((all_topk_vals.ravel(), all_topk_inds.ravel(), indptr), labels.shape)
            # remove padding if any
            if any(score_mat.indices == labels.shape[1]):
                score_mat.data[score_mat.indices == labels.shape[1]] = 0
                score_mat.eliminate_zeros()
            return score_mat


class DistAsymmetricLabelwiseAttentionNetwork(DistAsymmetricDualEncoderAll):
    def __init__(self, args, accelerator, data_manager):
        super().__init__(args, accelerator, data_manager)
        self.q_proj = nn.Linear(self.yembs.shape[-1], self.embs_dim)
        self.out_proj = nn.Linear(self.yembs.shape[-1], self.embs_dim) if args.out_proj else None
        self.ignore_cls = args.ignore_cls
    
    def _attn_block(self, x, query, attn_mask):
        x = F.scaled_dot_product_attention(query, x, x, attn_mask=attn_mask, dropout_p=(self.dropout.p if self.training else 0.0)) # (B, P, D) x (B, S, D) -> (B, P, D)
        return x
    
    def forward_head(self, all_xembs, attn_mask):
        if self.ignore_cls:
            all_xembs = all_xembs[:, 1:]
            attn_mask = attn_mask[:, 1:]
        query = self.q_proj(self.dropout(self.local_yembs)) # (L, D) -> (L, D)
        attn_mask = attn_mask.unsqueeze(1).bool() # (B, S) -> (B, 1, S)
        if self.out_proj is not None:
            output_yembs = F.normalize(self.out_proj(self.dropout(self.local_yembs)), dim=-1) # (L, D) -> (L, D)
        else:
            output_yembs = F.normalize(query, dim=-1)

        # forward in sub-batches
        logits = torch.zeros((all_xembs.shape[0], self.local_y_inds.shape[0]), device=self.get_device())
        for i in range(0, self.local_y_inds.shape[0], self.gc_bsz):
            input_range = range(i, min(i+self.gc_bsz, self.local_y_inds.shape[0]))
            query_sub = query[input_range].unsqueeze(0) # (L, D) -> (1, P, D)
            x = all_xembs
            x = self._attn_block(x, query_sub, attn_mask) # (B, P, D)
            x = F.normalize(x, dim=-1) # (B, P, D)
            output_yembs_sub = output_yembs[input_range].unsqueeze(0) # (L, D) -> (1, P, D)
            logits[:, input_range] = (output_yembs_sub * x).sum(dim=-1) # (1, P, D) x (B, P, D) -> (B, P)
        return logits


NETS = {
    'dist-clf-net': DistClassifierNetwork,
    'dist-lwan': DistLabelWiseAttentionNetwork,
    'dist-de-all': DistDualEncoderAll,
    'dist-ade-all': DistAsymmetricDualEncoder,
    'dist-asym-wde-all': DistAsymmetricDualEncoderAll,
    'dist-asym-zslwan': DistAsymmetricLabelwiseAttentionNetwork,
    }